/*
 * deviced-vibrator
 *
 * Copyright (c) 2016 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#include <stdio.h>
#include <stdbool.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <limits.h>
#include <fcntl.h>
#include <Ecore.h>
#include <sys/types.h>

#include "core/log.h"
#include "core/list.h"
#include "haptic.h"
#include "standard-vibcore.h"

#define CIRCLE_ON_PATH             "/sys/class/sec/motor/motor_on"
#define CIRCLE_OFF_PATH            "/sys/class/sec/motor/motor_off"

static int fd_play = -1, fd_stop = -1;
static dd_list *handle_list;
static Ecore_Timer *stop_timer;
static int unique_number;
static bool state = false;

static int stop_device(int device_handle);
static bool find_from_list(int handle)
{
	dd_list *elem;

	elem = DD_LIST_FIND(handle_list, (gpointer)(long)handle);
	if (elem)
		return true;

	return false;
}

static Eina_Bool timer_cb(void *data)
{
	int device_handle = (int)(long)data;
	_I("stop vibration by timer");

	/* stop previous vibration */
	stop_device(device_handle);

	return ECORE_CALLBACK_CANCEL;
}

static int get_device_count(int *count)
{
	/* suppose there is just one haptic device */
	if (count)
		*count = 1;

	return 0;
}

static int open_device(int device_index, int *device_handle)
{
	int n;
	bool found = false;
	dd_list *elem;

	if (!device_handle)
		return -EINVAL;

	/* if it is the first element */
	n = DD_LIST_LENGTH(handle_list);
	if (n == 0) {
		_I("Open");
		if (fd_play < 0) {
			fd_play = open(CIRCLE_ON_PATH, O_RDONLY);
			if (fd_play < 0) {
				_E("Failed to open %s : %d", CIRCLE_ON_PATH, errno);
				return -errno;
			}
		}
		if (fd_stop < 0) {
			fd_stop = open(CIRCLE_OFF_PATH, O_RDONLY);
			if (fd_stop < 0) {
				_E("Failed to open %s : %d", CIRCLE_OFF_PATH, errno);
				return -errno;
			}
		}
	}

	if (unique_number == INT_MAX)
		unique_number = 0;

	while (found != true) {
		++unique_number;
		elem = DD_LIST_FIND(handle_list, (gpointer)(long)unique_number);
		if (!elem)
			found = true;
	}

	/* add info to local list */
	DD_LIST_APPEND(handle_list, (gpointer)(long)unique_number);

	*device_handle = unique_number;
	return 0;
}

static int close_device(int device_handle)
{
	int r, n;
	bool found;

	found = find_from_list(device_handle);
	if (!found)
		return -EINVAL;

	if (fd_stop < 0) {
		fd_stop = open(CIRCLE_OFF_PATH, O_RDONLY);
		if (fd_stop < 0)
			return -ENODEV;
	}

	/* stop vibration */
	r = stop_device(device_handle);
	if (r < 0)
		_I("already stopped or failed to stop effect : %d", r);

	/* unregister existing timer */
	if (r >= 0) {
		_D("device handle %d is closed and timer deleted", device_handle);
		if (stop_timer) {
			ecore_timer_del(stop_timer);
			stop_timer = NULL;
		}
	}

	standard_vibrate_close();

	DD_LIST_REMOVE(handle_list, (gpointer)(long)device_handle);

	/* if it is the last element */
	n = DD_LIST_LENGTH(handle_list);
	if (n == 0) {
		_I("Close");
		if (fd_play > 0) {
			close(fd_play);
			fd_play = -1;
		}
		if (fd_stop > 0) {
			close(fd_stop);
			fd_stop = -1;
		}
	}

	return 0;
}

static int vibrate_monotone(int device_handle, int duration, int feedback, int priority, int *effect_handle)
{
	int ret;
	char buf[8];
	bool found;

	found = find_from_list(device_handle);
	if (!found)
		return -EINVAL;

	if (fd_play < 0) {
		fd_play = open(CIRCLE_ON_PATH, O_RDONLY);
		if (fd_play < 0)
			return -ENODEV;
	}

	/* Zero(0) is the infinitely vibration value */
	if (duration == HAPTIC_MODULE_DURATION_UNLIMITED)
		duration = 0;

	if (stop_timer)
		stop_device(device_handle);

	/* play vibration */
	ret = read(fd_play, buf, 8);
	if (ret < 0) {
		_E("failed to play");
		return -errno;
	}

	/* register timer */
	if (duration) {
		stop_timer = ecore_timer_add(duration/1000.f, timer_cb, (void *)(long)device_handle);
		if (!stop_timer)
			_E("Failed to add timer callback");
	}

	_D("device handle %d %dms", device_handle, duration);

	return 0;
}

static int vibrate_buffer(int device_handle, const unsigned char *vibe_buffer, int iteration, int feedback, int priority, int *effect_handle)
{
	_E("Not support feature");
	return -EACCES;
}

static int stop_device(int device_handle)
{
	int ret;
	char buf[8];
	bool found;

	found = find_from_list(device_handle);
	if (!found)
		return -EINVAL;

	if (fd_stop < 0) {
		fd_stop = open(CIRCLE_OFF_PATH, O_RDONLY);
		if (fd_stop < 0)
			return -ENODEV;
	}
	ret = read(fd_stop, buf, 8);
	if (ret < 0) {
		_E("failed to stop");
		return -errno;
	}
	if (stop_timer) {
		ecore_timer_del(stop_timer);
		stop_timer = NULL;
	}

	return 0;
}

static int get_device_state(int device_index, int *effect_state)
{
	if (!effect_state)
		return -EINVAL;

	*effect_state = state;
	return 0;
}

static int create_effect(unsigned char *vibe_buffer, int max_bufsize, haptic_module_effect_element *elem_arr, int max_elemcnt)
{
	_E("Not support feature");
	return -EACCES;
}

static int get_buffer_duration(int device_handle, const unsigned char *vibe_buffer, int *buffer_duration)
{
	_E("Not support feature");
	return -EACCES;
}

static int convert_binary(const unsigned char *vibe_buffer, int max_bufsize, const char *file_path)
{
	_E("Not support feature");
	return -EACCES;
}
/* END: Haptic Module APIs */

static const struct haptic_plugin_ops default_plugin = {
	.get_device_count    = get_device_count,
	.open_device         = open_device,
	.close_device        = close_device,
	.vibrate_monotone    = vibrate_monotone,
	.vibrate_buffer      = vibrate_buffer,
	.vibrate_effect      = standard_vibrate_effect,
	.is_supported        = standard_is_supported,
	.stop_device         = stop_device,
	.get_device_state    = get_device_state,
	.create_effect       = create_effect,
	.get_buffer_duration = get_buffer_duration,
	.convert_binary      = convert_binary,
};

static bool is_valid(void)
{
	int ret;

	if ((access(CIRCLE_ON_PATH, R_OK) != 0) ||
		(access(CIRCLE_OFF_PATH, R_OK) != 0)) {
		_E("Do not support wearable haptic device");
		state = false;
		return false;
	}

	ret = standard_config_parse();
	if (ret < 0)
		_E("failed to load standard vibration configuration file : %d", ret);

	state = true;
	_I("Support wearable haptic device");
	return true;
}

static const struct haptic_plugin_ops *load(void)
{
	standard_set_vib_function(&vibrate_monotone);
	return &default_plugin;
}

static const struct haptic_ops std_ops = {
	.type     = HAPTIC_STANDARD,
	.is_valid = is_valid,
	.load     = load,
};

HAPTIC_OPS_REGISTER(&std_ops)
